@@ -22,7 +22,6 @@
 #include "foot-features.h"
 #include "macros.h"
 #include "util.h"
-#include "version.h"
 #include "xmalloc.h"
 
 extern char **environ;
@@ -62,20 +61,6 @@ sendall(int sock, const void *_buf, size_t len)
     return len;
 }
 
-static const char *
-version_and_features(void)
-{
-    static char buf[256];
-    snprintf(buf, sizeof(buf),
-             "version: %s %cpgo %cime %cgraphemes %cassertions",
-             FOOT_VERSION,
-             feature_pgo() ? '+' : '-',
-             feature_ime() ? '+' : '-',
-             feature_graphemes() ? '+' : '-',
-             feature_assertions() ?  '+' : '-');
-    return buf;
-}
-
 static void
 print_usage(const char *prog_name)
 {
@@ -328,7 +313,7 @@ main(int argc, char *const *argv)
             break;
 
         case 'v':
-            printf("footclient %s\n", version_and_features());
+            print_version_and_features("footclient ");
             ret = EXIT_SUCCESS;
             goto err;
 
@@ -399,10 +384,10 @@ main(int argc, char *const *argv)
 
     const char *cwd = custom_cwd;
     if (cwd == NULL) {
-        errno = 0;
         size_t buf_len = 1024;
         do {
             _cwd = xrealloc(_cwd, buf_len);
+            errno = 0;
             if (getcwd(_cwd, buf_len) == NULL && errno != ERANGE) {
                 LOG_ERRNO("failed to get current working directory");
                 goto err;
