@@ -1,7 +1,8 @@
 #pragma once
 
-#include <stdint.h>
+#include <regex.h>
 #include <stdbool.h>
+#include <stdint.h>
 #include <uchar.h>
 
 #include <xkbcommon/xkbcommon.h>
@@ -27,7 +28,7 @@ struct font_size_adjustment {
     float percent;
 };
 
-enum cursor_style { CURSOR_BLOCK, CURSOR_UNDERLINE, CURSOR_BEAM };
+enum cursor_style { CURSOR_BLOCK, CURSOR_UNDERLINE, CURSOR_BEAM, CURSOR_HOLLOW };
 enum cursor_unfocused_style {
     CURSOR_UNFOCUSED_UNCHANGED,
     CURSOR_UNFOCUSED_HOLLOW,
@@ -60,6 +61,7 @@ enum binding_aux_type {
     BINDING_AUX_NONE,
     BINDING_AUX_PIPE,
     BINDING_AUX_TEXT,
+    BINDING_AUX_REGEX,
 };
 
 struct binding_aux {
@@ -73,6 +75,8 @@ struct binding_aux {
             uint8_t *data;
             size_t len;
         } text;
+
+        char *regex_name;
     };
 };
 
@@ -120,6 +124,13 @@ struct env_var {
 };
 typedef tll(struct env_var) env_var_list_t;
 
+struct custom_regex {
+    char *name;
+    char *regex;
+    regex_t preg;
+    struct config_spawn_template launch;
+};
+
 struct config {
     char *term;
     char *shell;
@@ -153,6 +164,9 @@ struct config {
     enum { STARTUP_WINDOWED, STARTUP_MAXIMIZED, STARTUP_FULLSCREEN } startup_mode;
 
     bool dpi_aware;
+    enum {GAMMA_CORRECT_DISABLED,
+          GAMMA_CORRECT_ENABLED,
+          GAMMA_CORRECT_AUTO} gamma_correct;
     struct config_font_list fonts[4];
     struct font_size_adjustment font_size_adjustment;
 
@@ -186,6 +200,7 @@ struct config {
         bool urgent;
         bool notify;
         bool flash;
+        bool system_bell;
         struct config_spawn_template command;
         bool command_focused;
     } bell;
@@ -219,12 +234,12 @@ struct config {
             OSC8_UNDERLINE_ALWAYS,
         } osc8_underline;
 
-        char32_t **protocols;
-        char32_t *uri_characters;
-        size_t prot_count;
-        size_t max_prot_len;
+        char *regex;
+        regex_t preg;
     } url;
 
+    tll(struct custom_regex) custom_regexes;
+
     struct {
         uint32_t fg;
         uint32_t bg;
@@ -385,6 +400,7 @@ struct config {
         bool box_drawing_solid_shades;
         bool font_monospace_warn;
         bool sixel;
+        enum { SHM_8_BIT, SHM_10_BIT } surface_bit_depth;
     } tweak;
 
     struct {
