@@ -16,6 +16,10 @@
 
 #define TIME_REFLOW 0
 
+#if defined(TIME_REFLOW)
+#include "misc.h"
+#endif
+
 /*
  * "sb" (scrollback relative) coordinates
  *
@@ -36,7 +40,8 @@ grid_row_abs_to_sb(const struct grid *grid, int screen_rows, int abs_row)
     return rebased_row;
 }
 
-int grid_row_sb_to_abs(const struct grid *grid, int screen_rows, int sb_rel_row)
+int
+grid_row_sb_to_abs(const struct grid *grid, int screen_rows, int sb_rel_row)
 {
     const int scrollback_start = grid->offset + screen_rows;
     int abs_row = sb_rel_row + scrollback_start;
@@ -434,7 +439,7 @@ grid_row_alloc(int cols, bool initialize)
 {
     struct row *row = xmalloc(sizeof(*row));
     row->dirty = false;
-    row->linebreak = false;
+    row->linebreak = true;
     row->extra = NULL;
     row->shell_integration.prompt_marker = false;
     row->shell_integration.cmd_start = -1;
@@ -496,7 +501,6 @@ grid_resize_without_reflow(
                sizeof(struct cell) * min(old_cols, new_cols));
 
         new_row->dirty = old_row->dirty;
-        new_row->linebreak = false;
         new_row->shell_integration.prompt_marker = old_row->shell_integration.prompt_marker;
         new_row->shell_integration.cmd_start = min(old_row->shell_integration.cmd_start, new_cols - 1);
         new_row->shell_integration.cmd_end = min(old_row->shell_integration.cmd_end, new_cols - 1);
@@ -707,7 +711,6 @@ _line_wrap(struct grid *old_grid, struct row **new_grid, struct row *row,
     } else {
         /* Scrollback is full, need to reuse a row */
         grid_row_reset_extra(new_row);
-        new_row->linebreak = false;
         new_row->shell_integration.prompt_marker = false;
         new_row->shell_integration.cmd_start = -1;
         new_row->shell_integration.cmd_end = -1;
@@ -811,7 +814,7 @@ tp_cmp(const void *_a, const void *_b)
 
 void
 grid_resize_and_reflow(
-    struct grid *grid, int new_rows, int new_cols,
+    struct grid *grid, const struct terminal *term, int new_rows, int new_cols,
     int old_screen_rows, int new_screen_rows,
     size_t tracking_points_count,
     struct coord *const _tracking_points[static tracking_points_count])
@@ -889,6 +892,8 @@ grid_resize_and_reflow(
                 i, tracking_points[i]->row, tracking_points[i]->col);
     }
 
+    int coalesced_linebreaks = 0;
+
     /*
      * Walk the old grid
      */
@@ -930,7 +935,8 @@ grid_resize_and_reflow(
 
         if (!old_row->linebreak && col_count > 0) {
             /* Don't truncate logical lines */
-            col_count = old_cols;
+            while (col_count < old_cols && old_row->cells[col_count].wc == 0)
+                col_count++;
         }
 
         xassert(col_count >= 0 && col_count <= old_cols);
@@ -954,7 +960,7 @@ grid_resize_and_reflow(
         /* Does this row have any URIs? */
         struct row_range *uri_range, *uri_range_terminator;
         struct row_range *underline_range, *underline_range_terminator;
-        struct row_data *extra = old_row->extra;
+        const struct row_data *extra = old_row->extra;
 
         if (extra != NULL && extra->uri_ranges.count > 0) {
             uri_range = &extra->uri_ranges.v[0];
@@ -978,213 +984,179 @@ grid_resize_and_reflow(
         } else
             underline_range = underline_range_terminator = NULL;
 
-        for (int start = 0, left = col_count; left > 0;) {
-            int end;
-            bool tp_break = false;
-            bool uri_break = false;
-            bool underline_break = false;
-            bool ftcs_break = false;
+        if (unlikely(col_count > 0 && coalesced_linebreaks > 0)) {
+            for (size_t apa = 0; apa < coalesced_linebreaks; apa++) {
+                /* Erase the remaining cells */
+                memset(&new_row->cells[new_col_idx], 0,
+                       (new_cols - new_col_idx) * sizeof(new_row->cells[0]));
+                new_row->linebreak = true;
 
-            /* Figure out where to end this chunk */
-            {
-                const int uri_col = uri_range != uri_range_terminator
-                    ? ((uri_range->start >= start ? uri_range->start : uri_range->end) + 1)
-                    : INT_MAX;
-                const int underline_col = underline_range != underline_range_terminator
-                    ? ((underline_range->start >= start ? underline_range->start : underline_range->end) + 1)
-                    : INT_MAX;
-                const int tp_col = tp != NULL ? tp->col + 1 : INT_MAX;
-                const int ftcs_col = old_row->shell_integration.cmd_start >= start
-                    ? old_row->shell_integration.cmd_start + 1
-                    : old_row->shell_integration.cmd_end >= start
-                    ? old_row->shell_integration.cmd_end + 1
-                    : INT_MAX;
-
-                end = min(col_count, min(min(tp_col, min(uri_col, underline_col)), ftcs_col));
-
-                uri_break = end == uri_col;
-                underline_break = end == underline_col;
-                tp_break = end == tp_col;
-                ftcs_break = end == ftcs_col;
+                if (r + 1 < old_rows)
+                    line_wrap();
             }
 
-            int cols = end - start;
-            xassert(cols > 0);
-            xassert(start + cols <= old_cols);
+            coalesced_linebreaks = 0;
+        }
+
+        for (int c = 0; c < col_count;) {
+            const struct cell *old = &old_row->cells[c];
+
+            /* Row full, emit newline and get a new, fresh, row */
+            xassert(new_col_idx <= new_cols);
+            if (unlikely(new_col_idx >= new_cols))
+                line_wrap();
+
+            char32_t wc = old->wc;
+            int width = 1;
+
+            if (unlikely(wc >= CELL_COMB_CHARS_LO && wc <= CELL_COMB_CHARS_HI)) {
+                const struct composed *composed =
+                    composed_lookup(term->composed, wc - CELL_COMB_CHARS_LO);
+
+                width = composed->forced_width > 0 ? composed->forced_width : composed->width;
+            } else if (unlikely(c + 1 < col_count && (old + 1)->wc >= CELL_SPACER + 1)) {
+                /* Wide character, get its width from the next cell's
+                   SPACER value */
+                width = (old + 1)->wc - CELL_SPACER + 1;
+            }
 
             /*
-             * Copy the row chunk to the new grid. Note that there may
-             * be fewer cells left on the new row than what we have in
-             * the chunk. I.e. the chunk may have to be split up into
-             * multiple memcpy:ies.
-             */
+             * Check if character fits, if not, emit spacers, and push
+               the character to the next row */
+            if (unlikely(new_col_idx + width > new_cols && width <= new_cols)) {
+                for (; new_col_idx < new_cols; new_col_idx++) {
+                    new_row->cells[new_col_idx].wc = CELL_SPACER;
+                    new_row->cells[new_col_idx].attrs = (struct attributes){0};
+                }
+                line_wrap();
+            }
 
-            for (int count = cols, from = start; count > 0;) {
-                xassert(new_col_idx <= new_cols);
-                int new_row_cells_left = new_cols - new_col_idx;
+            new_row->shell_integration.prompt_marker = old_row->shell_integration.prompt_marker;
 
-                /* Row full, emit newline and get a new, fresh, row */
-                if (new_row_cells_left <= 0) {
-                    line_wrap();
-                    new_row_cells_left = new_cols;
+            for (int i = 0; i < width; i++) {
+                if (unlikely(uri_range != NULL && uri_range != uri_range_terminator)) {
+                    if (unlikely(uri_range->start == c)) {
+                        reflow_range_start(
+                            uri_range, ROW_RANGE_URI, new_row, new_col_idx);
+                    }
+
+                    if (unlikely(uri_range->end == c)) {
+                        reflow_range_end(
+                            uri_range, ROW_RANGE_URI, new_row, new_col_idx);
+                        grid_row_uri_range_destroy(uri_range);
+                        uri_range++;
+                    }
                 }
 
-                /* Number of cells we can copy */
-                int amount = min(count, new_row_cells_left);
-                xassert(amount > 0);
+                if (unlikely(underline_range != NULL && underline_range != underline_range_terminator)) {
+                    if (unlikely(underline_range->start == c)) {
+                        reflow_range_start(
+                            underline_range, ROW_RANGE_UNDERLINE, new_row, new_col_idx);
+                    }
+
+                    if (unlikely(underline_range->end == c)) {
+                        reflow_range_end(
+                            underline_range, ROW_RANGE_UNDERLINE, new_row, new_col_idx);
+                        grid_row_underline_range_destroy(underline_range);
+                        underline_range++;
+                    }
+                }
+
+                if (unlikely(tp != NULL)) {
+                    if (unlikely(tp->col == c)) {
+                        do {
+                            xassert(tp->row == old_row_idx);
+
+                            tp->row = new_row_idx;
+                            tp->col = new_col_idx;
+
+                            next_tp++;
+                            tp = *next_tp;
+                        } while (tp->row == old_row_idx && tp->col == c);
+
+                        if (tp->row != old_row_idx)
+                            tp = NULL;
+
+                        LOG_DBG("next TP (tp=%p): %dx%d",
+                                (void*)tp, (*next_tp)->row, (*next_tp)->col);
+                    }
+                }
+
+                if (unlikely(old_row->shell_integration.cmd_start == c))
+                    new_row->shell_integration.cmd_start = new_col_idx;
+
+                if (unlikely(old_row->shell_integration.cmd_end == c))
+                    new_row->shell_integration.cmd_end = new_col_idx;
+
+                if (unlikely(width > new_cols)) {
+                    /* Wide character no longer fits on a row, replace
+                       it with a single space */
+                    new_row->cells[new_col_idx++].wc = 0;
+                    c++;
+
+                    /* Walk past the SPACER cells */
+                    for (int i = 1; i < width; i++, c++, old++)
+                        ;
+
+                    /* Continue with next character in the *old* grid */
+                    break;
+                }
+
+                new_row->cells[new_col_idx++] = *old;
 
                 /*
-                 * If we're going to reach the end of the new row, we
-                 * need to make sure we don't end in the middle of a
-                 * multi-column character.
+                 * TODO: simulate LCF instead?
+                 *
+                 * Rows have linebreak=true by default. This is needed
+                 * for a number of reasons. However, we want non-empty
+                 * rows to have linebreak=false, *until* we reach the
+                 * end of an old row with linebreak=true, at which
+                 * point we set linebreak=true on the new row.
                  */
-                int spacers = 0;
-                if (new_col_idx + amount >= new_cols) {
-                    /*
-                     * While the cell *after* the last cell is a CELL_SPACER
-                     *
-                     * This means we have a multi-column character
-                     * that doesn't fit on the current row. We need to
-                     * push it to the next row, and insert CELL_SPACER
-                     * cells as padding.
-                     */
-                    while (
-                        unlikely(
-                            amount > 1 &&
-                            from + amount < old_cols &&
-                            old_row->cells[from + amount].wc >= CELL_SPACER + 1))
-                    {
-                        amount--;
-                        spacers++;
-                    }
-
-                    xassert(
-                        amount == 1 ||
-                        old_row->cells[from + amount - 1].wc <= CELL_SPACER + 1);
-                }
-
-                xassert(new_col_idx + amount <= new_cols);
-                xassert(from + amount <= old_cols);
-
-                if (from == 0)
-                    new_row->shell_integration.prompt_marker = old_row->shell_integration.prompt_marker;
-
-                memcpy(
-                    &new_row->cells[new_col_idx], &old_row->cells[from],
-                    amount * sizeof(struct cell));
-
-                count -= amount;
-                from += amount;
-                new_col_idx += amount;
-
-                xassert(new_col_idx <= new_cols);
-
-                if (unlikely(spacers > 0)) {
-                    xassert(new_col_idx + spacers == new_cols);
-
-                    const struct cell *cell = &old_row->cells[from - 1];
-
-                    for (int i = 0; i < spacers; i++, new_col_idx++) {
-                        new_row->cells[new_col_idx].wc = CELL_SPACER;
-                        new_row->cells[new_col_idx].attrs = cell->attrs;
-                    }
-                }
+                new_row->linebreak = false;
+                old++;
+                c++;
             }
-
-            xassert(new_col_idx > 0);
-
-            if (tp_break) {
-                do {
-                    xassert(tp != NULL);
-                    xassert(tp->row == old_row_idx);
-                    xassert(tp->col == end - 1);
-
-                    tp->row = new_row_idx;
-                    tp->col = new_col_idx - 1;
-
-                    next_tp++;
-                    tp = *next_tp;
-                } while (tp->row == old_row_idx && tp->col == end - 1);
-
-                if (tp->row != old_row_idx)
-                    tp = NULL;
-
-                LOG_DBG("next TP (tp=%p): %dx%d",
-                        (void*)tp, (*next_tp)->row, (*next_tp)->col);
-            }
-
-            if (uri_break) {
-                xassert(uri_range != NULL);
-
-                if (uri_range->start == end - 1)
-                    reflow_range_start(
-                        uri_range, ROW_RANGE_URI, new_row, new_col_idx - 1);
-
-                if (uri_range->end == end - 1) {
-                    reflow_range_end(
-                        uri_range, ROW_RANGE_URI, new_row, new_col_idx - 1);
-                    grid_row_uri_range_destroy(uri_range);
-                    uri_range++;
-                }
-            }
-
-            if (underline_break) {
-                xassert(underline_range != NULL);
-
-                if (underline_range->start == end - 1)
-                    reflow_range_start(
-                        underline_range, ROW_RANGE_UNDERLINE, new_row, new_col_idx - 1);
-
-                if (underline_range->end == end - 1) {
-                    reflow_range_end(
-                        underline_range, ROW_RANGE_UNDERLINE, new_row, new_col_idx - 1);
-                    grid_row_underline_range_destroy(underline_range);
-                    underline_range++;
-                }
-            }
-
-            if (ftcs_break) {
-                xassert(old_row->shell_integration.cmd_start == start + cols - 1 ||
-                        old_row->shell_integration.cmd_end == start + cols - 1);
-
-                if (old_row->shell_integration.cmd_start == start + cols - 1)
-                    new_row->shell_integration.cmd_start = new_col_idx - 1;
-                if (old_row->shell_integration.cmd_end == start + cols - 1)
-                    new_row->shell_integration.cmd_end = new_col_idx - 1;
-            }
-
-            left -= cols;
-            start += cols;
         }
 
         if (old_row->linebreak) {
-            /* Erase the remaining cells */
-            memset(&new_row->cells[new_col_idx], 0,
-                   (new_cols - new_col_idx) * sizeof(new_row->cells[0]));
-            new_row->linebreak = true;
-
-            if (r + 1 < old_rows)
-                line_wrap();
-            else if (new_row->extra != NULL) {
-                if (new_row->extra->uri_ranges.count > 0) {
-                    /*
-                     * line_wrap() "closes" still-open URIs. Since
-                     * this is the *last* row, and since we're
-                     * line-breaking due to a hard line-break (rather
-                     * than running out of cells in the "new_row"),
-                     * there shouldn't be an open URI (it would have
-                     * been closed when we reached the end of the URI
-                     * while reflowing the last "old" row).
-                     */
-                    int last_idx = new_row->extra->uri_ranges.count - 1;
-                    xassert(new_row->extra->uri_ranges.v[last_idx].end >= 0);
-                }
-
-                if (new_row->extra->underline_ranges.count > 0) {
-                    int last_idx = new_row->extra->underline_ranges.count - 1;
-                    xassert(new_row->extra->underline_ranges.v[last_idx].end >= 0);
+            if (col_count > 0) {
+                /* Erase the remaining cells */
+                memset(&new_row->cells[new_col_idx], 0,
+                       (new_cols - new_col_idx) * sizeof(new_row->cells[0]));
+                new_row->linebreak = true;
 
+                if (r + 1 < old_rows) {
+                    /* Not the last (old) row */
+                    line_wrap();
+                } else if (new_row->extra != NULL) {
+                    if (new_row->extra->uri_ranges.count > 0) {
+                        /*
+                         * line_wrap() "closes" still-open URIs. Since
+                         * this is the *last* row, and since we're
+                         * line-breaking due to a hard line-break (rather
+                         * than running out of cells in the "new_row"),
+                         * there shouldn't be an open URI (it would have
+                         * been closed when we reached the end of the URI
+                         * while reflowing the last "old" row).
+                         */
+                        int last_idx = new_row->extra->uri_ranges.count - 1;
+                        xassert(new_row->extra->uri_ranges.v[last_idx].end >= 0);
+                    }
+
+                    if (new_row->extra->underline_ranges.count > 0) {
+                        int last_idx = new_row->extra->underline_ranges.count - 1;
+                        xassert(new_row->extra->underline_ranges.v[last_idx].end >= 0);
+                    }
                 }
+            } else {
+                /*
+                 * rows have linebreak=true by default. But we don't
+                 * want trailing empty lines to result in actual lines
+                 * in the new grid (think: empty window with prompt at
+                 * the top)
+                 */
+                coalesced_linebreaks++;
             }
         }
 
@@ -1274,15 +1246,26 @@ grid_resize_and_reflow(
     saved_cursor.row = min(saved_cursor.row, new_screen_rows - 1);
     saved_cursor.col = min(saved_cursor.col, new_cols - 1);
 
+    if (grid->cursor.lcf) {
+        if (cursor.col + 1 < new_cols) {
+            cursor.col++;
+            grid->cursor.lcf = false;
+        }
+    }
+
+    if (grid->saved_cursor.lcf) {
+        if (saved_cursor.col + 1 < new_cols) {
+            saved_cursor.col++;
+            grid->saved_cursor.lcf = false;
+        }
+    }
+
     grid->cur_row = new_grid[(grid->offset + cursor.row) & (new_rows - 1)];
     xassert(grid->cur_row != NULL);
 
     grid->cursor.point = cursor;
     grid->saved_cursor.point = saved_cursor;
 
-    grid->cursor.lcf = false;
-    grid->saved_cursor.lcf = false;
-
     /* Free sixels we failed to "map" to the new grid */
     tll_foreach(untranslated_sixels, it)
         sixel_destroy(&it->item);
