@@ -31,7 +31,6 @@
 #include "shm.h"
 #include "terminal.h"
 #include "util.h"
-#include "version.h"
 #include "xmalloc.h"
 #include "xsnprintf.h"
 
@@ -46,21 +45,6 @@ fdm_sigint(struct fdm *fdm, int signo, void *data)
     return true;
 }
 
-static const char *
-version_and_features(void)
-{
-    static char buf[256];
-    snprintf(buf, sizeof(buf),
-             "version: %s %cpgo %cime %cgraphemes %ctoplevel-icon %cassertions",
-             FOOT_VERSION,
-             feature_pgo() ? '+' : '-',
-             feature_ime() ? '+' : '-',
-             feature_graphemes() ? '+' : '-',
-             feature_xdg_toplevel_icon() ? '+' : '-',
-             feature_assertions() ? '+' : '-');
-    return buf;
-}
-
 static void
 print_usage(const char *prog_name)
 {
@@ -378,7 +362,7 @@ main(int argc, char *const *argv)
             break;
 
         case 'v':
-            printf("foot %s\n", version_and_features());
+            print_version_and_features("foot ");
             return EXIT_SUCCESS;
 
         case 'h':
@@ -406,7 +390,7 @@ main(int argc, char *const *argv)
         argv += optind;
     }
 
-    LOG_INFO("%s", version_and_features());
+    LOG_INFO("%s", version_and_features);
 
     {
         struct utsname name;
@@ -516,7 +500,6 @@ main(int argc, char *const *argv)
         (enum fcft_log_colorize)log_colorize,
         as_server && log_syslog,
         (enum fcft_log_class)log_level);
-    fcft_set_scaling_filter(conf.tweak.fcft_filter);
 
     if (conf_server_socket_path != NULL) {
         free(conf.server_socket_path);
@@ -551,10 +534,10 @@ main(int argc, char *const *argv)
     char *_cwd = NULL;
 
     if (cwd == NULL) {
-        errno = 0;
         size_t buf_len = 1024;
         do {
             _cwd = xrealloc(_cwd, buf_len);
+            errno = 0;
             if (getcwd(_cwd, buf_len) == NULL && errno != ERANGE) {
                 LOG_ERRNO("failed to get current working directory");
                 goto out;
