@@ -31,6 +31,8 @@ bool render_xcursor_set(
     struct seat *seat, struct terminal *term, enum cursor_shape shape);
 bool render_xcursor_is_valid(const struct seat *seat, const char *cursor);
 
+void render_overlay(struct terminal *term);
+
 struct render_worker_context {
     int my_id;
     struct terminal *term;
@@ -45,3 +47,5 @@ struct csd_data {
 };
 
 struct csd_data get_csd_data(const struct terminal *term, enum csd_surface surf_idx);
+
+bool render_do_linear_blending(const struct terminal *term);
