@@ -9,6 +9,7 @@
 #include <xkbcommon/xkbcommon.h>
 
 /* Wayland protocols */
+#include <color-management-v1.h>
 #include <fractional-scale-v1.h>
 #include <presentation-time.h>
 #include <primary-selection-unstable-v1.h>
@@ -19,10 +20,8 @@
 #include <xdg-decoration-unstable-v1.h>
 #include <xdg-output-unstable-v1.h>
 #include <xdg-shell.h>
-
-#if defined(HAVE_XDG_TOPLEVEL_ICON)
- #include <xdg-toplevel-icon-v1.h>
-#endif
+#include <xdg-system-bell-v1.h>
+#include <xdg-toplevel-icon-v1.h>
 
 #include <fcft/fcft.h>
 #include <tllist.h>
@@ -57,6 +56,7 @@ enum touch_state {
 struct wayl_surface {
     struct wl_surface *surf;
     struct wp_viewport *viewport;
+    struct wp_color_management_surface_v1 *color_management;
 };
 
 struct wayl_sub_surface {
@@ -447,9 +447,18 @@ struct wayland {
 
     struct wp_single_pixel_buffer_manager_v1 *single_pixel_manager;
 
-#if defined(HAVE_XDG_TOPLEVEL_ICON)
     struct xdg_toplevel_icon_manager_v1 *toplevel_icon_manager;
-#endif
+
+    struct xdg_system_bell_v1 *system_bell;
+
+    struct {
+        struct wp_color_manager_v1 *manager;
+        struct wp_image_description_v1 *img_description;
+        bool have_intent_perceptual;
+        bool have_feat_parametric;
+        bool have_tf_ext_linear;
+        bool have_primaries_srgb;
+    } color_management;
 
     bool presentation_timings;
     struct wp_presentation *presentation;
@@ -466,6 +475,11 @@ struct wayland {
 
     /* WL_SHM >= 2 */
     bool use_shm_release;
+
+    bool shm_have_argb2101010:1;
+    bool shm_have_xrgb2101010:1;
+    bool shm_have_abgr2101010:1;
+    bool shm_have_xbgr2101010:1;
 };
 
 struct wayland *wayl_init(
@@ -492,6 +506,7 @@ void wayl_win_destroy(struct wl_window *win);
 void wayl_win_scale(struct wl_window *win, const struct buffer *buf);
 void wayl_win_alpha_changed(struct wl_window *win);
 bool wayl_win_set_urgent(struct wl_window *win);
+bool wayl_win_ring_bell(const struct wl_window *win);
 
 bool wayl_win_csd_titlebar_visible(const struct wl_window *win);
 bool wayl_win_csd_borders_visible(const struct wl_window *win);
